#!/usr/bin/env python3
import time
import io

import unittest
from joule import api
import folder, stream, module, data


async def setup():
    node = api.Node()
    """
    live
        sum1
        filt1
    arch
        sum1
        filt1
        free:
            stream1
            stream2
        """


def main():
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    suite.addTests(loader.loadTestsFromModule(folder))
    suite.addTests(loader.loadTestsFromModule(stream))
    suite.addTests(loader.loadTestsFromModule(module))
    suite.addTests(loader.loadTestsFromModule(data))
    output = io.StringIO()
    runner = unittest.TextTestRunner(stream=output, failfast=True)
    result = runner.run(suite)
    if result.wasSuccessful():
        print("OK")
        return 0
    else:
        print("FAIL")
        output.seek(0)
        print(output.read())
        return -1


if __name__ == '__main__':
    time.sleep(1)
    exit(main())



#!/usr/bin/env python3

import joule.client
import asyncio


class Adder(joule.client.FilterModule):
    """ Add DC offset to input """
    
    def custom_args(self, parser):
        parser.add_argument("offset", type=int, default=0,
                            help="apply an offset")
        
    async def run(self, parsed_args, inputs, outputs):
        stream_in: joule = inputs["input"]
        stream_out = outputs["output"]
        while not self.stop_requested:
            try:
                sarray = await stream_in.read()
                sarray["data"] += parsed_args.offset
                await asyncio.sleep(0.25)
                await stream_out.write(sarray)
                stream_in.consume(len(sarray))
                if stream_in.end_of_interval:
                    print("closing interval")
                    await stream_out.close_interval()

            except joule.EmptyPipe:
                print("got an empty pipe, exiting")
                exit(1)
            

if __name__ == "__main__":
    r = Adder()
    r.start()



# NOTE: Removed top level imports to reduce module load time

#from joule.client.reader_module import ReaderModule
#from joule.client.filter_module import FilterModule
#from joule.client.composite_module import CompositeModule
#from joule.client.base_module import BaseModule
#from joule.models.pipes import LocalPipe, EmptyPipe, InputPipe, OutputPipe, Pipe
#from joule.models import DataStream, Element
#from joule import utilities



import click
#from .copy import data_copy
#from .read import cmd as data_read
#from .remove import data_remove
#from .intervals import intervals
#from .consolidate import consolidate
#from .merge import merge
#from .mean import mean
#from .median import median
#from .ingest import ingest
from joule.cli.lazy_group import LazyGroup

@click.group(name="data",
             cls=LazyGroup,
             lazy_subcommands={"copy": "joule.cli.data.copy.data_copy",
                               "read": "joule.cli.data.read.cmd",
                               "delete": "joule.cli.data.delete.data_delete",
                               "intervals": "joule.cli.data.intervals.intervals",
                               "consolidate": "joule.cli.data.consolidate.consolidate",
                               "merge": "joule.cli.data.merge.merge",
                               "ingest": "joule.cli.data.ingest.ingest"})
def data():
    """Interact with data streams."""
    pass  # pragma: no cover

@click.group(name="filter",
             cls=LazyGroup,
             lazy_subcommands={"mean": "joule.cli.data.mean",
                               "median": "joule.cli.data.median"})
def filter():
    """Filter stream data."""
    pass  # pragma: no cover

data.add_command(filter)

"""
data.add_command(data_copy)
data.add_command(merge)
data.add_command(data_read)
data.add_command(data_remove)
data.add_command(intervals)
data.add_command(consolidate)
data.add_command(ingest)

filter.add_command(mean)
filter.add_command(median)
"""


from sqlalchemy.orm import Session
from aiohttp import web
import numpy as np
import asyncio
import logging
from joule import app_keys

from joule.models import (folder, DataStore, DataStream,
                          InsufficientDecimationError, DataError,
                          pipes)
from joule.models.supervisor import Supervisor

from joule.errors import SubscriptionError

log = logging.getLogger('joule')


async def read_json(request: web.Request):
    return await read(request, json=True)


async def read(request: web.Request, json=False):
    if 'subscribe' in request.query and request.query['subscribe'] == '1':
        return await _subscribe(request, json)
    else:
        return await _read(request, json)


async def _read(request: web.Request, json):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.get(DataStream,request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)

    # parse optional parameters
    params = {'start': None, 'end': None, 'max-rows': None, 'decimation-level': None}
    param = ""  # to appease type checker

    try:
        for param in params:
            if param in request.query:
                params[param] = int(request.query[param])
    except ValueError:
        return web.Response(text="parameter [%s] must be an int" % param, status=400)

    # make sure parameters make sense
    if ((params['start'] is not None and params['end'] is not None) and
            (params['start'] >= params['end'])):
        return web.Response(text="[start] must be < [end]", status=400)
    if params['max-rows'] is not None and params['max-rows'] <= 0:
        return web.Response(text="[max-rows] must be > 0", status=400)
    if params['decimation-level'] is not None and params['decimation-level'] <= 0:
        return web.Response(text="[decimation-level] must be > 0", status=400)

    # --- Binary Streaming Handler ---
    resp = None

    async def stream_data(data: np.ndarray, layout, factor):
        nonlocal resp

        if resp is None:
            resp = web.StreamResponse(status=200,
                                      headers={'joule-layout': layout,
                                               'joule-decimation': str(factor)})
            resp.enable_chunked_encoding()
            await resp.prepare(request)
        await resp.write(data.tobytes())

    # --- JSON Handler ---

    data_blocks = []  # array of data segments
    data_segment = None
    decimation_factor = 1

    async def retrieve_data(data: np.ndarray, layout, factor):
        nonlocal data_blocks, data_segment, decimation_factor
        decimation_factor = factor
        if np.array_equal(data, pipes.interval_token(layout)):
            if data_segment is not None:
                data_blocks.append(data_segment.tolist())
                data_segment = None
        else:
            data = np.c_[data['timestamp'][:, None], data['data']]

            if data_segment is None:
                data_segment = data
            else:
                data_segment = np.vstack((data_segment, data))

    if json:
        callback = retrieve_data
    else:
        callback = stream_data

    # create an extraction task
    try:
        await data_store.extract(stream, params['start'], params['end'],
                                 callback=callback,
                                 max_rows=params['max-rows'],
                                 decimation_level=params['decimation-level'])
    except InsufficientDecimationError as e:
        return web.Response(text="decimated data is not available: %s" % e, status=400)
    except DataError as e:
        msg = str(e)
        if 'no such stream' in msg.lower() and (params['decimation-level'] is not None):  # pragma: no cover
            # clean up error message when user requested a particular decimation level
            msg = "requested decimation level [%d] does not exist" % params['decimation-level']
        return web.Response(text="read error: %s" % msg, status=400)

    if json:
        # put the last data_segment on
        if data_segment is not None:
            data_blocks.append(data_segment.tolist())
        return web.json_response({"data": data_blocks, "decimation_factor": decimation_factor})
    else:
        if resp is None:
            return web.json_response(text="this stream has no data", status=400)
        return resp


async def _subscribe(request: web.Request, json: bool):
    db: Session = request.app[app_keys.db]
    supervisor: Supervisor = request.app[app_keys.supervisor]
    if json:
        return web.Response(text="JSON subscription not implemented", status=400)

    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.get(DataStream,request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)
    pipe = pipes.LocalPipe(stream.layout)
    try:
        unsubscribe = supervisor.subscribe(stream, pipe)
    except SubscriptionError:
        return web.Response(text="stream is not being produced", status=400)
    resp = web.StreamResponse(status=200,
                              headers={'joule-layout': stream.layout,
                                       'joule-decimation': '1'})
    resp.enable_chunked_encoding()

    try:
        await resp.prepare(request)
    except ConnectionResetError:
        unsubscribe()
        return resp

    try:
        while True:
            try:
                data = await pipe.read()
            except pipes.EmptyPipe:
                unsubscribe()
                return resp
            pipe.consume(len(data))
            if len(data) > 0:
                await resp.write(data.tobytes())
            if pipe.end_of_interval:
                await resp.write(pipes.interval_token(stream.layout).tobytes())
    except asyncio.CancelledError as e:
        unsubscribe()
        # propogate the CancelledError up
        raise e
    except ConnectionResetError:
        unsubscribe()
        return resp


async def intervals(request: web.Request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.get(DataStream,request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)
    # parse time bounds if specified
    try:
        if 'start' in request.query:
            start = int(request.query['start'])
        else:
            start = None
        if 'end' in request.query:
            end = int(request.query['end'])
        else:
            end = None
    except ValueError:
        return web.Response(text="[start] and [end] must be an integers", status=400)

    # make sure parameters make sense
    if (start is not None and end is not None) and start >= end:
        return web.Response(text="[start] must be < [end]", status=400)

    return web.json_response(await data_store.intervals(stream, start, end))


async def write(request: web.Request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.get(DataStream,request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)
    # spawn in inserter task
    stream.is_destination = True
    db.commit()

    pipe = pipes.InputPipe(name="inbound", stream=stream, reader=request.content)
    if 'merge-gap' in request.query:
        try:
            merge_gap = int(request.query['merge-gap'])
        except ValueError:
            return web.Response(text="merge-gap must be an integer", status=400)
        
        if merge_gap < 0:
            return web.Response(text="merge-gap must be >= 0", status=400)
    else:
        merge_gap = 0
    try:
        task = await data_store.spawn_inserter(stream, pipe, insert_period=0, merge_gap=merge_gap)
        await task
    except DataError as e:
        stream.is_destination = False
        db.commit()
        print("closing stream due to error")
        return web.Response(text=str(e), status=400)
    except asyncio.CancelledError as e:
        raise e
    finally:
        stream.is_destination = False
        db.commit()
    return web.Response(text="ok")


async def remove(request: web.Request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.get(DataStream,request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)

    # parse time bounds
    start = None
    end = None
    try:
        if 'start' in request.query:
            start = int(request.query['start'])
        if 'end' in request.query:
            end = int(request.query['end'])
    except ValueError:
        return web.Response(text="[start] and [end] must be integers", status=400)

    # make sure bounds make sense
    if ((start is not None and end is not None) and
            (start >= end)):
        return web.Response(text="[start] must be < [end]", status=400)

    await data_store.remove(stream, start, end)
    return web.Response(text="ok")


async def consolidate(request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.get(DataStream,request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)

    # parse time bounds
    start = None
    end = None
    try:
        if 'start' in request.query:
            start = int(request.query['start'])
        if 'end' in request.query:
            end = int(request.query['end'])
    except ValueError:
        return web.Response(text="[start] and [end] must be integers", status=400)

    # make sure bounds make sense
    if ((start is not None and end is not None) and
            (start >= end)):
        return web.Response(text="[start] must be < [end]", status=400)

    # parse the max_gap parameter
    if 'max_gap' not in request.query:
        return web.Response(text="specify max_gap as us integer", status=400)
    try:
        max_gap = int(request.query['max_gap'])
        if max_gap <= 0:
            raise ValueError()
    except ValueError:
        return web.Response(text="max_gap must be postive integer", status=400)
    num_removed = await data_store.consolidate(stream, start, end, max_gap)
    return web.json_response(data={"num_consolidated": num_removed})


async def decimate(request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.get(DataStream,request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)
    await data_store.decimate(stream)
    return web.Response(text="ok")

async def drop_decimations(request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    # find the requested stream
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.get(DataStream, request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if stream is None:
        return web.Response(text="stream does not exist", status=404)
    await data_store.drop_decimations(stream)
    return web.Response(text="ok")



#!/usr/bin/python3 -u

"""
Run the configuration in follower
"""

import os
import sys
import subprocess
import time
import shlex
import socket
import signal

SOURCE_DIR = "/joule"
MODULE_SCRIPT_DIR = "/joule/tests/e2e/module_scripts"
JOULE_CONF_DIR = "/joule/tests/e2e/follower"


def prep_system():
    os.symlink(MODULE_SCRIPT_DIR, "/module_scripts")


def run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL):
    return subprocess.run(shlex.split(cmd), stdout=stdout, stderr=stderr)


def main():
    prep_system()

    config_file = os.path.join(JOULE_CONF_DIR, "main.conf")
    subprocess.run(("joule admin erase --yes --config %s" % config_file).split(" "))

    # add a key entry for the master node
    # get API permissions
    os.environ["LOGNAME"] = "e2e"
    os.environ["JOULE_USER_CONFIG_DIR"] = "/tmp/joule_user"
    with open(os.devnull, 'w') as devnull:
        subprocess.run(("joule admin authorize --config %s" % config_file).split(" "),
                       stdout=devnull)

    jouled = subprocess.Popen(["jouled", "--config", config_file],
                              stdout=sys.stdout,
                              stderr=sys.stderr,
                              universal_newlines=True)
    time.sleep(3)
    # follow node1.joule
    subprocess.run("joule master add joule https://node1.joule:8088".split(" "))  # ,

    # this will just hang, node1 exits and terminates this container
    stdout, _ = jouled.communicate()
    print(stdout)
    print(_)
    return 0


if __name__ == "__main__":
    exit(main())



from aiohttp.test_utils import unused_port
import os
import warnings
import unittest

from joule.models import DataStream, Element, pipes
from joule.models.data_store.nilmdb import NilmdbStore, bytes_per_row
from joule.models.data_store.errors import InsufficientDecimationError, DataError
from tests.models.data_store.fake_nilmdb import FakeNilmdb, FakeStream
from tests import helpers

STREAM_LIST = os.path.join(os.path.dirname(__file__), 'stream_list.json')
warnings.simplefilter('always')


class TestNilmdbStore(unittest.IsolatedAsyncioTestCase):
    use_default_loop = False
    forbid_get_event_loop = True

    async def asyncSetUp(self):
        self.fake_nilmdb = FakeNilmdb()
        url = await self.fake_nilmdb.start()

        # use a 0 insert period for test execution
        self.store = NilmdbStore(url, 0, 60)

        # make a couple example streams
        # stream1 int8_3
        self.stream1 = DataStream(id=1, name="stream1", datatype=DataStream.DATATYPE.INT8,
                                  elements=[Element(name="e%d" % x) for x in range(3)])

        # stream2 uint16_4
        self.stream2 = DataStream(id=2, name="stream2", datatype=DataStream.DATATYPE.UINT16,
                                  elements=[Element(name="e%d" % x) for x in range(4)])

    async def asyncTearDown(self) -> None:
        await self.fake_nilmdb.stop()
        await self.store.close()

    async def test_initialize(self):

        await self.store.initialize([self.stream1, self.stream2])
        await self.store.close()
        # can initialize streams multiple times
        await self.store.initialize([self.stream1])

        self.assertEqual(len(self.fake_nilmdb.posts), 3)
        self.assertEqual(len(self.fake_nilmdb.streams), 2)
        self.assertEqual(self.fake_nilmdb.streams["/joule/1"].layout, "int8_3")
        self.assertEqual(self.fake_nilmdb.streams["/joule/2"].layout, "uint16_4")

    async def test_initialize_errors(self):
        # when nilmdb returns invalid error data
        self.fake_nilmdb.stub_stream_create = True
        self.fake_nilmdb.response = 'notjson'
        self.fake_nilmdb.http_code = 500

        with self.assertRaises(DataError) as e:
            await self.store.initialize([self.stream1, self.stream2])
        await self.store.close()
        # when nilmdb server is not available
        self.store.server = 'http://localhost:%d' % unused_port()
        with self.assertRaises(DataError):
            await self.store.initialize([self.stream1, self.stream2])

    async def test_insert(self):
        await self.store.initialize([self.stream1])
        nrows = 300
        data = helpers.create_data(layout="int8_3", length=nrows)
        # first insert
        await self.store.insert(self.stream1, data['timestamp'][0],
                                data['timestamp'][-1] + 1, data)
        self.assertEqual(self.fake_nilmdb.streams["/joule/1"].rows, nrows)
        # another insert
        data = helpers.create_data(layout="int8_3", start=data['timestamp'][-1] + 1,
                                   length=nrows)
        await self.store.insert(self.stream1, data['timestamp'][0],
                                data['timestamp'][-1] + 1, data)
        self.assertEqual(self.fake_nilmdb.streams["/joule/1"].rows, nrows * 2)

    async def test_insert_error_on_overlapping_data(self):
        # when nilmdb server rejects the data (eg overlapping timestamps)
        await self.store.initialize([self.stream1])
        self.fake_nilmdb.streams['/joule/1'] = FakeStream(
            layout=self.stream1.layout, start=0, end=500, rows=500)
        data = helpers.create_data(layout="int8_3", start=25, step=1, length=20)
        with self.assertRaises(DataError)as e:
            await self.store.insert(self.stream1, 25, 1000, data)
        self.assertTrue('overlaps' in str(e.exception))

    async def test_extract(self):
        await self.store.initialize([])
        # initialize nilmdb state
        nrows = 1000
        self.fake_nilmdb.streams['/joule/1'] = FakeStream(
            layout=self.stream1.layout, start=100, end=1000, rows=nrows)
        self.fake_nilmdb.streams['/joule/1~decim-4'] = FakeStream(
            layout=self.stream1.decimated_layout, start=100, end=1000, rows=nrows // 4)
        self.fake_nilmdb.streams['/joule/1~decim-16'] = FakeStream(
            layout=self.stream1.decimated_layout, start=100, end=1000, rows=nrows // 16)
        self.fake_nilmdb.streams['/joule/1~decim-64'] = FakeStream(
            layout=self.stream1.decimated_layout, start=100, end=1000, rows=nrows // 64)

        extracted_data = []

        async def callback(data, layout, decimated):
            extracted_data.append(data)

        # should return *raw* data when max_rows and decimation_level are None
        await self.store.extract(self.stream1, start=100, end=1000, callback=callback)
        self.assertEqual(len(self.fake_nilmdb.extract_calls), 1)
        path = self.fake_nilmdb.extract_calls[0]['path']
        self.assertTrue(path, '/joule/1')
        rcvd_rows = 0
        last_row = []
        for chunk in extracted_data:
            if chunk is not None:
                rcvd_rows += len(chunk)
                last_row = chunk[-1]
        self.assertEqual(nrows, rcvd_rows)

        # should return *decimated* data when max_rows!=None
        self.fake_nilmdb.extract_calls = []  # reset
        extracted_data = []
        await self.store.extract(self.stream1, start=100, end=1000,
                                 callback=callback, max_rows=int(1.5 * (nrows / 16)))
        # expect count req from raw and decim-16, and extract from data from decim-16
        self.assertEqual(len(self.fake_nilmdb.extract_calls), 3)
        path = self.fake_nilmdb.extract_calls[0]['path']
        self.assertTrue(path, '/joule/1~decim-16')
        rcvd_rows = 0
        last_row = []
        for chunk in extracted_data:
            if chunk is not None:
                rcvd_rows += len(chunk)
                last_row = chunk[-1]
        self.assertEqual(nrows // 16, rcvd_rows)

        # should raise exception if data is not decimated enough
        with self.assertRaises(InsufficientDecimationError):
            await self.store.extract(self.stream1, start=100, end=1000,
                                     callback=callback, max_rows=nrows // 128)

        # should return data from a specified decimation level
        self.fake_nilmdb.extract_calls = []  # reset
        extracted_data = []
        await self.store.extract(self.stream1, start=100, end=1000,
                                 callback=callback, decimation_level=64)
        # expect extract from decim-64
        self.assertEqual(len(self.fake_nilmdb.extract_calls), 1)
        path = self.fake_nilmdb.extract_calls[0]['path']
        self.assertTrue(path, '/joule/1~decim-64')
        rcvd_rows = 0
        last_row = []
        for chunk in extracted_data:
            if chunk is not None:
                rcvd_rows += len(chunk)
                last_row = chunk[-1]
        self.assertEqual(nrows // 64, rcvd_rows)

        # if max_rows and decimation_level specified, error if
        # too much data
        with self.assertRaises(InsufficientDecimationError):
            await self.store.extract(self.stream1, start=100, end=1000,
                                     callback=callback, decimation_level=64,
                                     max_rows=nrows // 128)

        # error if decimation_level does not exist
        with self.assertRaises(DataError):
            await self.store.extract(self.stream1, start=100, end=1000,
                                     callback=callback, decimation_level=5)

    async def test_extract_empty_data(self):
        await self.store.initialize([])
        # initialize nilmdb state
        self.fake_nilmdb.streams['/joule/1'] = FakeStream(
            layout=self.stream1.layout, start=100, end=1000, rows=0)

        async def callback(data, layout, decimation_factor):
            self.assertEqual(decimation_factor, 1)
            self.assertEqual(self.stream1.layout, layout)
            self.assertEqual(pipes.compute_dtype(self.stream1.layout), data.dtype)
            if len(data) == 0:
                self.assertEqual(0, len(data['timestamp']))
            else:
                for row in data:
                    self.assertEqual(row, pipes.interval_token(self.stream1.layout))

        # should return empty data with the correct type when there is no data
        await self.store.extract(self.stream1, start=100, end=1000, max_rows=100, callback=callback)
        await self.store.extract(self.stream1, start=100, end=1000, callback=callback)
        await self.store.extract(self.stream1, start=100, end=1000, callback=callback)

    async def test_extract_errors(self):
        await self.store.initialize([])
        # initialize nilmdb state
        nrows = 1000
        self.fake_nilmdb.streams['/joule/1'] = FakeStream(
            layout=self.stream1.layout, start=100, end=1000, rows=nrows)
        self.fake_nilmdb.streams['/joule/1~decim-4'] = FakeStream(
            layout=self.stream1.decimated_layout, start=100, end=1000, rows=nrows // 4)
        # level 16 has no data
        self.fake_nilmdb.streams['/joule/1~decim-16'] = FakeStream(
            layout=self.stream1.decimated_layout, start=100, end=1000, rows=0)

        extracted_data = []

        async def callback(data, layout, decimated):
            extracted_data.append(data)

        # returns error because required level is emptyF
        with self.assertRaises(InsufficientDecimationError) as e:
            await self.store.extract(self.stream1, start=100, end=1000, callback=callback,
                                     max_rows=(nrows // 4 - 10))
        self.assertTrue('empty' in str(e.exception))

    async def test_intervals(self):
        await self.store.initialize([])
        actual_intervals = [[1, 5],
                            [10, 15], [15, 18], [18, 20],
                            [30, 31], [31, 40]]
        consolidated_intervals = [[1, 5], [10, 20], [30, 40]]
        self.fake_nilmdb.streams['/joule/1'] = FakeStream(
            layout=self.stream1.layout, start=100, end=1000,
            intervals=actual_intervals)
        rcvd_intervals = await self.store.intervals(self.stream1,
                                                    start=None,
                                                    end=None)
        self.assertEqual(consolidated_intervals, rcvd_intervals)

        # handles empty intervals
        self.fake_nilmdb.streams['/joule/1'].intervals = None
        rcvd_intervals = await self.store.intervals(self.stream1,
                                                    start=None,
                                                    end=None)
        self.assertEqual([], rcvd_intervals)

        # ignore 404 error if stream does not exist (return empty interval)
        new_stream = helpers.create_stream("new", "float32_3", id=100)
        rcvd_intervals = await self.store.intervals(new_stream,
                                                    start=None,
                                                    end=None)
        self.assertEqual([], rcvd_intervals)

        # propogate any other error
        self.fake_nilmdb.stub_stream_intervals = True
        self.fake_nilmdb.response = 'notjson'
        self.fake_nilmdb.http_code = 500
        with self.assertRaises(DataError):
            await self.store.intervals(self.stream1,
                                       start=None,
                                       end=None)

    async def test_remove(self):
        await self.store.initialize([])
        # initialize nilmdb state
        self.fake_nilmdb.streams['/joule/1'] = FakeStream(
            layout=self.stream1.layout, start=1, end=100, rows=100)
        self.fake_nilmdb.streams['/joule/1~decim-4'] = FakeStream(
            layout=self.stream1.decimated_layout, start=1, end=100, rows=100)
        self.fake_nilmdb.streams['/joule/1~decim-16'] = FakeStream(
            layout=self.stream1.decimated_layout, start=1, end=100, rows=100)
        self.fake_nilmdb.streams['/joule/1~decim-64'] = FakeStream(
            layout=self.stream1.decimated_layout, start=1, end=100, rows=100)

        # removes data between time bounds at all decimation levels
        await self.store.remove(self.stream1, start=1, end=1000)
        self.assertEqual(len(self.fake_nilmdb.remove_calls), 4)
        paths = ['/joule/1', '/joule/1~decim-4',
                 '/joule/1~decim-16', '/joule/1~decim-64']
        for call in self.fake_nilmdb.remove_calls:
            self.assertTrue(call["path"] in paths)

    async def test_info(self):
        await self.store.initialize([])
        with open(STREAM_LIST, 'r') as f:
            self.fake_nilmdb.stream_list_response = f.read()
        streams = [helpers.create_stream('S%d' % x, 'float32_3', id=x) for x in range(8)]
        info = await self.store.info(streams)
        # streams 1, 2, 6, 7, 8 are in the json file but 8 is not in the streams request
        for stream_id in [1, 2, 6, 7]:
            self.assertTrue(stream_id in info)
        # S1 should be empty
        s1_info = info[1]
        self.assertIsNone(s1_info.start)
        self.assertIsNone(s1_info.end)
        self.assertEqual(s1_info.rows, 0)
        self.assertEqual(s1_info.bytes, 0)
        # S2 should have the following values
        s2_info = info[2]
        self.assertEqual(s2_info.start, 10)
        self.assertEqual(s2_info.end, 100)
        self.assertEqual(s2_info.rows, 64)
        self.assertEqual(s2_info.bytes, 1628)
        self.assertEqual(s2_info.total_time, 100)

    async def test_bytes_per_row(self):
        await self.store.initialize([])
        examples = [['float32_8', 40],
                    ['uint8_1', 9],
                    ['int16_20', 48],
                    ['float64_1', 16]]
        for example in examples:
            self.assertEqual(bytes_per_row(example[0]), example[1])
        bad_layouts = ['float32', '', 'abc', 'other32_8']
        for layout in bad_layouts:
            with self.assertRaises(ValueError) as e:
                bytes_per_row(layout)
            self.assertTrue(layout in str(e.exception))

    async def test_dbinfo(self):
        await self.store.initialize([])
        dbinfo = await self.store.dbinfo()
        # constants in fake_nilmdb
        self.assertEqual(dbinfo.path, '/opt/data')
        self.assertEqual(dbinfo.free, 178200829952)

    async def test_destroy(self):
        await self.store.initialize([])
        # initialize nilmdb state
        self.fake_nilmdb.streams['/joule/1'] = FakeStream(
            layout=self.stream1.layout, start=1, end=100, rows=100)
        self.fake_nilmdb.streams['/joule/1~decim-4'] = FakeStream(
            layout=self.stream1.decimated_layout, start=1, end=100, rows=100)
        self.fake_nilmdb.streams['/joule/1~decim-16'] = FakeStream(
            layout=self.stream1.decimated_layout, start=1, end=100, rows=100)
        self.fake_nilmdb.streams['/joule/1~decim-64'] = FakeStream(
            layout=self.stream1.decimated_layout, start=1, end=100, rows=100)

        # removes data between time bounds at all decimation levels
        await self.store.destroy(self.stream1)
        self.assertEqual(len(self.fake_nilmdb.remove_calls), 4)
        paths = ['/joule/1', '/joule/1~decim-4',
                 '/joule/1~decim-16', '/joule/1~decim-64']
        for call in self.fake_nilmdb.remove_calls:
            self.assertTrue(call["path"] in paths)
        # and destroys the stream
        for call in self.fake_nilmdb.destroy_calls:
            self.assertTrue(call["path"] in paths)



from click.testing import CliRunner
from aiohttp.test_utils import unused_port
import warnings
import asyncio
import numpy as np
import multiprocessing
import time
import os
import signal
import json
import psutil
import datetime
import unittest

from ..fake_joule import FakeJoule, FakeJouleTestCase
from tests.models.data_store.fake_nilmdb import FakeNilmdb, FakeStream
from joule.cli import main
from joule.models import DataStream, Element, StreamInfo
from tests import helpers

warnings.simplefilter('always')


class TestDataCopyWithNilmDb(FakeJouleTestCase):

    def _start_nilmdb(self, msgs):
        port = unused_port()
        self.nilmdb_proc = multiprocessing.Process(target=self._run_nilmdb, args=(port, msgs))
        self.nilmdb_proc.start()
        ready = False
        time.sleep(0.01)

        while not ready:
            for conn in psutil.net_connections():
                if conn.laddr.port == port:
                    ready = True
                    break
            else:
                time.sleep(0.01)
        return "http://127.0.0.1:%d" % port

    def _stop_nilmdb(self):
        if self.nilmdb_proc is None:
            return
        # aiohttp doesn't always quit with SIGTERM
        os.kill(self.nilmdb_proc.pid, signal.SIGKILL)
        # join any zombies
        self.nilmdb_proc.join()
        del self.nilmdb_proc

    def _run_nilmdb(self, port, msgs):

        loop = asyncio.new_event_loop()
        server = FakeNilmdb()
        # create an empty stream
        server.streams['/existing/float32_3'] = FakeStream(
            layout='float32_3', start=0, end=0, rows=0)
        # create a source stream
        server.streams['/test/source'] = FakeStream(
            layout='float32_3', start=0, end=100, rows=1000)

        self.nilmdb_task = loop.create_task(server.start(port, msgs))
        loop.run_forever()

    def test_copies_data_to_nilmdb(self):

        source_server = FakeJoule()
        nilmdb_msgs = multiprocessing.Queue()
        dest_url = self._start_nilmdb(nilmdb_msgs)
        # create the source and destination streams
        src_data = create_source_data(source_server)  # helpers.create_data(src.layout)

        src_stream = DataStream(id=1, name="dest", keep_us=100,
                                datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.datetime.utcnow())
        src_stream.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in
                               range(3)]
        source_server.add_stream('/test/destination', src_stream, StreamInfo(None, None, 0), None)

        self.start_server(source_server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy',
                                      '-d', dest_url, '/test/source', '/existing/float32_3'],
                               input='y\n', catch_exceptions=True)
        _print_result_on_error(result)
        # expect data transfer call
        nilmdb_call = nilmdb_msgs.get()
        self.assertEqual('stream_insert', nilmdb_call['action'])
        data = nilmdb_call['data']
        np.testing.assert_array_equal(src_data, data)
        self.assertEqual(0, result.exit_code)
        self._stop_nilmdb()
        self.stop_server()
        del nilmdb_msgs

    def test_creates_nilmdb_stream_if_necessary(self):
        source_server = FakeJoule()
        nilmdb_msgs = multiprocessing.Queue()
        dest_url = self._start_nilmdb(nilmdb_msgs)
        # create just the source stream
        src_data = create_source_data(source_server)  # helpers.create_data(src.layout)

        src_stream = DataStream(id=1, name="dest", keep_us=100,
                                datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.datetime.utcnow())
        src_stream.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in
                               range(3)]
        source_server.add_stream('/test/destination', src_stream, StreamInfo(None, None, 0), None)

        self.start_server(source_server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy',
                                      '-d', dest_url, '/test/source', '/test/destination'], catch_exceptions=False)
        _print_result_on_error(result)
        # expect a stream create call
        nilmdb_call = nilmdb_msgs.get()
        self.assertEqual('stream_create', nilmdb_call['action'])
        self.assertEqual({'path': '/test/destination', 'layout': 'float32_3'}, nilmdb_call['params'])
        # expect a metadata call
        nilmdb_call = nilmdb_msgs.get()
        self.assertEqual('set_metadata', nilmdb_call['action'])
        self.assertEqual('/test/destination', nilmdb_call['params']['path'])
        self.assertEqual('config_key__', list(json.loads(nilmdb_call['params']['data']).keys())[0])
        # expect data transfer call
        nilmdb_call = nilmdb_msgs.get()
        self.assertEqual('stream_insert', nilmdb_call['action'])
        data = nilmdb_call['data']
        np.testing.assert_array_equal(src_data, data)
        self.assertEqual(0, result.exit_code)
        self._stop_nilmdb()
        self.stop_server()
        del nilmdb_msgs

    def test_copies_data_from_nilmdb(self):
        dest_server = FakeJoule()
        nilmdb_msgs = multiprocessing.Queue()
        source_url = self._start_nilmdb(nilmdb_msgs)
        # create the source and destination streams

        self.start_server(dest_server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'copy',
                                      '--source-url', source_url,
                                      '-d', "fake_joule", '/test/source', '/test/destination'],
                               catch_exceptions=False)
        _print_result_on_error(result)
        self.assertEqual(0, result.exit_code)
        mock_entry = self.msgs.get()
        self.assertEqual(len(mock_entry.data), 1000)
        self._stop_nilmdb()
        self.stop_server()
        del nilmdb_msgs


def _print_result_on_error(result):
    if result.exit_code == -1:
        print("exception: ", result.exception)
    if result.exit_code != 0:
        print("output: ", result.output)


def create_source_data(server):
    # create the source stream
    src = DataStream(id=0, name="source", keep_us=100,
                     datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.datetime.utcnow())
    src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]

    # source has 100 rows of data
    src_data = helpers.create_data(src.layout, length=100, start=0, step=1)
    ts = src_data['timestamp']
    intervals = [[ts[0], ts[99]]]
    src_info = StreamInfo(int(src_data['timestamp'][0]), int(src_data['timestamp'][-1]),
                          len(src_data))
    server.add_stream('/test/source', src, src_info, src_data, intervals)
    return src_data



from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from typing import List, Dict, Tuple
import numpy as np
import asyncio
import argparse
import psycopg2
import testing.postgresql

from unittest import mock
from typing import Optional, Callable, Coroutine

from joule.models.data_store.errors import DataError, InsufficientDecimationError
from joule.models.data_store.event_store import EventStore
from joule.models.data_store.event_store import StreamInfo as EventStreamInfo
from joule.models import (Base, DataStore, DataStream,
                          StreamInfo, DbInfo, pipes,
                          worker, EventStream)
from joule.errors import SubscriptionError
from joule.services import parse_pipe_config
from tests import helpers

Loop = asyncio.AbstractEventLoop


def create_db(pipe_configs: List[str]) -> Tuple[Session, testing.postgresql.Postgresql]:
    postgresql = testing.postgresql.Postgresql()
    db_url = postgresql.url()
    conn = psycopg2.connect(db_url)
    cur = conn.cursor()
    cur.execute("CREATE SCHEMA metadata")
    conn.commit()
    conn.close()
    engine = create_engine(db_url)
    Base.metadata.create_all(engine)
    db = Session(bind=engine)
    for pipe_config in pipe_configs:
        parse_pipe_config.run(pipe_config, db)

    db.commit()
    return db, postgresql


class MockEventStore(EventStore):
    def __init__(self):
        self.stream_info = {}
        self.destroyed_stream_id = None

    async def initialize(self):
        pass

    async def info(self, streams: List[EventStream]) -> Dict[int, EventStreamInfo]:
        info_dict = {}
        for s in streams:
            if s in self.stream_info:
                info_dict[s.id] = self.stream_info[s]
        return info_dict

    async def destroy(self, stream: EventStream):
        self.destroyed_stream_id = stream.id


class MockStore(DataStore):
    def __init__(self):
        self.stream_info = {}
        # stubs to track data controller execution
        self.nchunks = 3
        self.nintervals = 1
        self.raise_data_error = False
        self.raise_decimation_error = False
        self.no_data = False
        self.raise_data_error = False
        self.inserted_data = False
        self.removed_data_bounds = (None, None)
        self.destroyed_stream_id = None
        self.raise_data_error = False
        self.dropped_decimations = False
        self.redecimated_data = False
        self.merge_gap = None

    async def initialize(self, streams: List[DataStream]):
        pass

    async def insert(self, stream: DataStream,
                     data: np.ndarray, start: int, end: int):
        pass

    async def spawn_inserter(self, stream: DataStream, pipe: pipes.InputPipe, insert_period=None, merge_gap=0) -> asyncio.Task:
        async def task():
            if self.raise_data_error:
                raise DataError("test error")
            self.inserted_data = True
            self.merge_gap = merge_gap

        return asyncio.create_task(task())

    def configure_extract(self, nchunks, nintervals=1,
                          decimation_error=False,
                          data_error=False,
                          no_data=False):
        self.nchunks = nchunks
        self.nintervals = nintervals
        self.raise_decimation_error = decimation_error
        self.raise_data_error = data_error
        self.no_data = no_data

    async def extract(self, stream: DataStream, start: Optional[int], end: Optional[int],
                      callback: Callable[[np.ndarray, str, bool], Coroutine],
                      max_rows: int = None, decimation_level=1):
        if self.no_data:
            return  # do not call the callback func
        if self.raise_data_error:
            raise DataError("nilmdb error")
        if self.raise_decimation_error:
            raise InsufficientDecimationError("insufficient decimation")
        if decimation_level is not None and decimation_level > 1:
            layout = stream.decimated_layout
        else:
            decimation_level = 1
            layout = stream.layout
        for i in range(self.nintervals):
            for x in range(self.nchunks):
                await callback(helpers.create_data(layout, length=25), layout, decimation_level)
            if i < (self.nintervals - 1):
                await callback(pipes.interval_token(layout), layout, decimation_level)

    async def intervals(self, stream: 'DataStream', start: Optional[int], end: Optional[int]):
        if start is None:
            if end is None:
                # just a random set of intervals
                return [[0, 100], [200, 300]]
            else:
                return [[end - 10, end]]
        else:
            if end is None:
                return [[start, start + 10]]
            else:
                return [[start, end]]

    async def drop_decimations(self, stream: 'DataStream'):
        self.dropped_decimations = True

    async def decimate(self, stream: 'DataStream'):
        self.redecimated_data = True

    async def remove(self, stream: DataStream, start: Optional[int], end: Optional[int]):
        self.removed_data_bounds = (start, end)

    async def destroy(self, stream: DataStream):
        self.destroyed_stream_id = stream.id

    async def destroy_all(self):
        raise Exception("not implemented!")

    async def info(self, streams: List[DataStream]) -> Dict[int, StreamInfo]:
        info_dict = {}
        for s in streams:
            if s in self.stream_info:
                info_dict[s.id] = self.stream_info[s]
        return info_dict

    async def dbinfo(self) -> DbInfo:
        return DbInfo('/file/path', 0, 0, 0, 0)

    async def consolidate(self, stream: 'DataStream', start: Optional[int], end: Optional[int], max_gap: int) -> int:
        return 0

    def close(self):
        pass

    # -- special mock tools --
    def set_info(self, stream: DataStream, info: StreamInfo):
        self.stream_info[stream] = info


class MockWorker:
    def __init__(self, name, inputs, outputs, uuid=1, is_app=False, socket=None):
        self.name = name
        self.description = "description for %s" % name
        self.is_app = is_app
        self.uuid = uuid
        self.interface_socket = socket
        self.input_connections = []
        for (name, path) in inputs.items():
            self.input_connections.append(argparse.Namespace(name=name, location=path))
        self.output_connections = []
        for (name, path) in outputs.items():
            self.output_connections.append(argparse.Namespace(name=name, location=path))

    async def statistics(self) -> worker.Statistics:
        return worker.Statistics(100, 100, 1.0, 100)

    @property
    def logs(self):
        return ["log entry1", "log entry2"]


class MockSupervisor:
    def __init__(self):
        self.subscription_pipe = None
        self.subscribed_stream = None
        self.raise_error = False
        # if True do not close the pipe (simulate an ongoing transaction)
        self.hang_pipe = False
        self.unsubscribe_mock = mock.Mock()
        self.unsubscribe_calls = 0

    def unsubscribe(self):
        self.unsubscribe_calls += 1

    def subscribe(self, stream: DataStream, pipe: pipes.LocalPipe):
        if self.raise_error:
            raise SubscriptionError()

        self.subscribed_stream = stream
        while True:
            try:
                data = self.subscription_pipe.read_nowait()
                self.subscription_pipe.consume(len(data))
                pipe.write_nowait(data)
            except pipes.EmptyPipe:
                if not self.hang_pipe:
                    pipe.close_nowait()
                break
            if self.subscription_pipe.end_of_interval:
                pipe.close_interval_nowait()
        # return a mock as the unsubscribe callback
        return self.unsubscribe



import testing.postgresql
import numpy as np
import asyncpg
import datetime
import itertools
import tempfile
import shutil
import os
import sys
import unittest

from joule.models import DataStream, Element, pipes
from joule.models.data_store import psql_helpers
from joule.models.data_store.timescale import TimescaleStore
from tests import helpers
from tests.models.pipes.reader import QueueReader

SQL_DIR = os.path.join(os.path.dirname(__file__), '..', '..', '..', 'src', 'joule', 'sql')


class TestTimescale(unittest.IsolatedAsyncioTestCase):
    use_default_loop = False
    forbid_get_event_loop = True

    async def asyncSetUp(self):
        # set up the pscql database
        self.psql_dir = tempfile.TemporaryDirectory()
        self.postgresql = testing.postgresql.Postgresql(base_dir=self.psql_dir.name)
        self.postgresql.stop()
        # now that the directory structure is created, customize the *.conf file
        src = os.path.join(os.path.dirname(__file__), "postgresql.conf")
        dest = os.path.join(self.psql_dir.name, "data", "postgresql.conf")
        shutil.copyfile(src, dest)
        # restart the database
        self.postgresql = testing.postgresql.Postgresql(base_dir=self.psql_dir.name)

        self.db_url = self.postgresql.url()
        # self.db_url = "postgresql://joule:joule@127.0.0.1:5432/joule"
        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
        await conn.execute("CREATE EXTENSION IF NOT EXISTS timescaledb CASCADE ")
        await conn.execute("CREATE USER joule_module")
        # load custom functions
        for file in os.listdir(SQL_DIR):
            file_path = os.path.join(SQL_DIR, file)
            with open(file_path, 'r') as f:
                await conn.execute(f.read())
        # await conn.execute("GRANT EXECUTE ON ALL FUNCTIONS IN SCHEMA public TO joule_module;")
        # await conn.execute("DROP SCHEMA IF EXISTS data CASCADE")
        # await conn.execute("CREATE SCHEMA data")
        # await conn.execute("GRANT ALL ON SCHEMA data TO public")
        await conn.close()

    async def asyncTearDown(self):
        self.postgresql.stop()
        self.psql_dir.cleanup()

    async def test_runner(self):
        tests = [
            self._test_merge_gaps_on_write,
            self._test_basic_insert_extract,
                 self._test_extract_data_with_intervals,
                 self._test_extract_decimated_data,
                 self._test_db_info,
                 self._test_info,
                 self._test_intervals,
                 self._test_remove,
                 self._test_destroy,
                 self._test_row_count,
                 self._test_actions_on_empty_streams,
                 self._test_consolidate,
                 self._test_consolidate_with_time_bounds,
                 self._test_remove_decimations,
                 self._test_redecimates_data]
        for test in tests:
            conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
            await conn.execute("DROP SCHEMA IF EXISTS data CASCADE")
            await conn.execute("CREATE SCHEMA data")
            await conn.execute("GRANT ALL ON SCHEMA data TO public")

            self.store = TimescaleStore(self.db_url, 0, 60,)
            await self.store.initialize([])
            # make a sample stream with data
            self.test_stream = DataStream(id=100, name="stream1", datatype=DataStream.DATATYPE.FLOAT32,
                                          keep_us=DataStream.KEEP_ALL, decimate=True,
                                          elements=[Element(name="e%d" % x) for x in range(3)])
            pipe = pipes.LocalPipe(self.test_stream.layout)
            self.test_data = helpers.create_data(layout=self.test_stream.layout, length=1005)
            task = self.store.spawn_inserter(self.test_stream, pipe)
            await pipe.write(self.test_data)
            await pipe.close()
            runner = await task
            await runner
            await conn.close()
            # await self.store.initialize([])
            await test()
            # simulate the nose2 test output
            sys.stdout.write('o')
            await self.store.close()
            sys.stdout.flush()

    async def _test_basic_insert_extract(self):
        stream_id = 990
        self.store.extract_block_size = 500
        psql_types = ['double precision', 'real', 'bigint', 'integer', 'smallint']
        datatypes = [DataStream.DATATYPE.FLOAT64, DataStream.DATATYPE.FLOAT32, DataStream.DATATYPE.INT64,
                     DataStream.DATATYPE.INT32, DataStream.DATATYPE.INT16]
        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
        for i in range(len(datatypes)):
            datatype = datatypes[i]
            psql_type = psql_types[i]
            for n_elements in range(1, 5):
                test_stream = DataStream(id=stream_id, name="stream1", datatype=datatype, keep_us=DataStream.KEEP_ALL,
                                         elements=[Element(name="e%d" % x) for x in range(n_elements)])
                test_stream.decimate = True
                source = QueueReader()
                pipe = pipes.InputPipe(stream=test_stream, reader=source)
                nrows = 955
                data = helpers.create_data(layout=test_stream.layout, length=nrows)
                task = await self.store.spawn_inserter(test_stream, pipe)
                for chunk in helpers.to_chunks(data, 300):
                    await source.put(chunk.tobytes())
                await task

                # make sure the correct tables have been created
                records = await conn.fetch('''SELECT table_name FROM information_schema.tables 
                                                    WHERE table_schema='data';''')
                tables = list(itertools.chain(*records))
                for table in ['stream%d' % stream_id, 'stream%d_intervals' % stream_id]:
                    self.assertIn(table, tables)

                # check the column data types
                records = await conn.fetch('''SELECT column_name, data_type FROM information_schema.columns 
                                                            WHERE table_name='stream%d' AND table_schema='data';''' % stream_id)
                (names, types) = zip(*records)
                expected_elements = ['time'] + ['elem%d' % x for x in range(n_elements)]
                self.assertCountEqual(names, expected_elements)
                expected_psql_types = tuple(['timestamp without time zone'] + [psql_type for x in range(n_elements)])
                self.assertEqual(types, expected_psql_types)
                self.assertEqual(len(records), n_elements + 1)

                # extract raw data
                extracted_data = []

                async def callback(rx_data, layout, factor):
                    self.assertEqual(layout, test_stream.layout)
                    self.assertEqual(factor, 1)
                    extracted_data.append(rx_data)
                
                await self.store.extract(test_stream, start=None, end=None, callback=callback)
                extracted_data = np.hstack(extracted_data)
                np.testing.assert_array_equal(extracted_data, data)

                level = 64
                data_mean = np.mean(extracted_data['data'][:level], axis=0)
                data_max = np.max(extracted_data['data'][:level], axis=0)
                data_min = np.min(extracted_data['data'][:level], axis=0)

                # extract decimated data
                async def d_callback(rx_data, layout, factor):
                    self.assertEqual(layout, test_stream.decimated_layout)
                    self.assertEqual(factor, level)
                    extracted_data.append(rx_data)

                extracted_data = []
                await self.store.extract(test_stream, decimation_level=level,
                                         start=None, end=None, callback=d_callback)
                extracted_data = np.hstack(extracted_data)
                expected = np.hstack((data_mean, data_min, data_max))
                np.testing.assert_array_almost_equal(expected, extracted_data['data'][0])
                stream_id += 1
        await conn.close()

    async def _test_extract_data_with_intervals(self):
        test_stream = DataStream(id=1, name="stream1", datatype=DataStream.DATATYPE.FLOAT32, keep_us=DataStream.KEEP_ALL,
                                 decimate=True, elements=[Element(name="e%d" % x) for x in range(3)])
        pipe = pipes.LocalPipe(test_stream.layout)
        nrows = 955
        data = helpers.create_data(layout=test_stream.layout, length=nrows)
        task = await self.store.spawn_inserter(test_stream, pipe)
        for chunk in helpers.to_chunks(data, 300):
            await pipe.write(chunk)
            await pipe.close_interval()
        await pipe.close()
        await task

        # extract data
        extracted_data = []

        async def callback(rx_data, layout, factor):
            self.assertEqual(layout, test_stream.layout)
            self.assertEqual(factor, 1)
            extracted_data.append(rx_data)

        await self.store.extract(test_stream, start=None, end=None, callback=callback)
        extracted_data = np.hstack(extracted_data)
        # check for interval boundaries
        np.testing.assert_array_equal(extracted_data[300], pipes.interval_token(test_stream.layout))
        np.testing.assert_array_equal(extracted_data[601], pipes.interval_token(test_stream.layout))
        np.testing.assert_array_equal(extracted_data[902], pipes.interval_token(test_stream.layout))

    async def _test_merge_gaps_on_write(self):
        # make sure successive writes less than a merge_gap apart are merged
        test_stream = DataStream(id=1, name="stream1", datatype=DataStream.DATATYPE.FLOAT32, keep_us=DataStream.KEEP_ALL,
                                 decimate=True, elements=[Element(name="e%d" % x) for x in range(3)])
        nrows = 300
        data = helpers.create_data(layout=test_stream.layout, length=nrows, step=1000, start=15e6)
        data_1 = data[:100]
        data_2 = data[100:200]
        data_3 = data[200:]
        # write the first chunk
        pipe = pipes.LocalPipe(test_stream.layout)
        task = await self.store.spawn_inserter(test_stream, pipe)
        await pipe.write(data_1)
        await pipe.close()
        await task

        # write the second chunk inside the merge_gap setting
        pipe = pipes.LocalPipe(test_stream.layout)
        task = await self.store.spawn_inserter(test_stream, pipe, merge_gap=1100)
        await pipe.write(data_2)
        await pipe.close()
        await task

        # write the third chunk outside the merge_gap setting
        pipe = pipes.LocalPipe(test_stream.layout)
        task = await self.store.spawn_inserter(test_stream, pipe, merge_gap=900)
        await pipe.write(data_3)
        await pipe.close()
        await task

        # extract data
        extracted_data = []

        intervals = await self.store.intervals(test_stream,start=None,end=None)
        #print(intervals)
        async def callback(rx_data, layout, factor):
            self.assertEqual(layout, test_stream.layout)
            self.assertEqual(factor, 1)
            extracted_data.append(rx_data)

        await self.store.extract(test_stream, start=None, end=None, callback=callback)
        extracted_data = np.hstack(extracted_data)
        self.assertEqual(len(extracted_data),301)
        # check for interval boundaries
        for i in range(200):
            np.testing.assert_array_equal(extracted_data[i], data[i])
        np.testing.assert_array_equal(extracted_data[200], pipes.interval_token(test_stream.layout))
        for i in range(201, 301):
            np.testing.assert_array_equal(extracted_data[i], data[i-1])
       

    async def _test_extract_decimated_data(self):

        extracted_data = []

        async def callback(rx_data, layout, factor):
            self.assertEqual(layout, self.test_stream.decimated_layout)
            self.assertEqual(factor, 4)
            extracted_data.append(rx_data)

        await self.store.extract(self.test_stream, max_rows=16,
                                 start=self.test_data['timestamp'][66],
                                 end=self.test_data['timestamp'][64 + 4 * 16],
                                 callback=callback)
        extracted_data = np.hstack(extracted_data)
        self.assertLessEqual(len(extracted_data), 16)
        self.assertGreater(len(extracted_data), 4)

    async def _test_remove_decimations(self):

        # make sure the stream is decimated:
        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
        table_names = await psql_helpers.get_decimation_table_names(conn, self.test_stream, with_schema=False)
        for table in [f'stream100_{4**i}' for i in range(1,6)]:
           self.assertIn(table, table_names)
        await self.store.drop_decimations(self.test_stream)
        # drop the decimations
        table_names = await psql_helpers.get_decimation_table_names(conn, self.test_stream, with_schema=False)
        # make sure the tables are dropped
        self.assertEqual(len(table_names), 0)
        await conn.close()

    async def _test_redecimates_data(self):
        # create decimated data
        test_stream = DataStream(id=1, name="stream1", datatype=DataStream.DATATYPE.FLOAT32,
                                 keep_us=DataStream.KEEP_ALL,
                                 elements=[Element(name="e%d" % x) for x in range(3)])
        test_stream.decimate = True
        source = QueueReader()
        pipe = pipes.InputPipe(stream=test_stream, reader=source)
        nrows = 1024
        data = helpers.create_data(layout=test_stream.layout, length=nrows)
        task = await self.store.spawn_inserter(test_stream, pipe)
        for chunk in helpers.to_chunks(data, 16):
            await source.put(chunk.tobytes())
            await source.put(pipes.interval_token(test_stream.layout).tobytes())
        await task
        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
        table_names = await psql_helpers.get_decimation_table_names(conn, test_stream, with_schema=False)
        # interval breaks cause decimation to stop after level 16 (level 64 is created but empty)
        self.assertEqual(len(table_names),3)
        orig_data = await conn.fetch("SELECT * FROM data.stream1_16")

        # redecimate doesn't change anything
        await self.store.decimate(test_stream)
        redecimated_data = await conn.fetch("SELECT * FROM data.stream1_16")
        self.assertEqual(orig_data, redecimated_data)

        # now consolidate the intervals
        await self.store.consolidate(test_stream, start=None, end=None,
                                     max_gap = data['timestamp'][1]-data['timestamp'][0])
        # redecimate
        await self.store.decimate(test_stream)
        table_names = await psql_helpers.get_decimation_table_names(conn, test_stream, with_schema=False)
        for table in [f'stream1_{4**i}' for i in range(1,6)]:
           self.assertIn(table, table_names)
        await conn.close()

    async def _test_nondecimating_inserter(self):
        # TODO
        pass

    async def _test_row_count(self):
        test_data = helpers.create_data(layout=self.test_stream.layout, length=10000)
        test_stream = DataStream(id=95, name="stream1", datatype=DataStream.DATATYPE.FLOAT32, keep_us=DataStream.KEEP_ALL,
                                 decimate=True, elements=[Element(name="e%d" % x) for x in range(3)])
        pipe = pipes.LocalPipe(test_stream.layout)
        task = await self.store.spawn_inserter(test_stream, pipe)
        await pipe.write(test_data)
        await pipe.close()
        await task
        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)
        # test to make sure nrows is within 10% of actual value
        # Test: [start, end]
        nrows = await psql_helpers.get_row_count(conn, test_stream,
                                                 None,
                                                 None)
        self.assertGreater(nrows, len(test_data) * 0.9)
        # Test: [ts,end]
        nrows = await psql_helpers.get_row_count(conn, test_stream, test_data[len(test_data) // 2][0], None)
        self.assertLess(abs(nrows - len(test_data) // 2), 0.1 * len(test_data))
        # Test: [start, ts]
        nrows = await psql_helpers.get_row_count(conn, test_stream,
                                                 None,
                                                 test_data[len(test_data) // 3][0])
        self.assertLess(abs(nrows - len(test_data) // 3), 0.1 * len(test_data))

        # Test: [ts, ts]
        nrows = await psql_helpers.get_row_count(conn, test_stream,
                                                 test_data[2 * len(test_data) // 6][0],
                                                 test_data[3 * len(test_data) // 6][0])
        self.assertLess(abs(nrows - len(test_data) // 6), 0.1 * len(test_data))

        # Test: [ts, ts] (no data)
        nrows = await psql_helpers.get_row_count(conn, test_stream,
                                                 test_data['timestamp'][0] - 100,
                                                 test_data['timestamp'][0] - 50)
        self.assertEqual(0, nrows)

        # Test row count for stream with no data tables
        empty_stream = DataStream(id=96, name="empty", datatype=DataStream.DATATYPE.FLOAT64,
                                  keep_us=100, decimate=True, elements=[Element(name="e%d" % x) for x in range(8)])
        nrows = await psql_helpers.get_row_count(conn, empty_stream,
                                                 None,
                                                 None)
        self.assertEqual(0, nrows)
        nrows = await psql_helpers.get_row_count(conn, empty_stream,
                                                 test_data[len(test_data) // 2][0],
                                                 None)
        self.assertEqual(0, nrows)
        nrows = await psql_helpers.get_row_count(conn, test_stream,
                                                 test_data['timestamp'][0] - 100,
                                                 test_data['timestamp'][0] - 50)
        self.assertEqual(0, nrows)
        await conn.close()

    async def _test_remove(self):

        #  XXXXXXX------XXXXXXX
        #        ^|     ^
        # remove a chunk of data from the middle
        await self.store.remove(self.test_stream,
                                self.test_data['timestamp'][300],
                                self.test_data['timestamp'][400])
        # XXXXXXX------===XXXX
        #       ^|        ^
        # remove another chunk
        await self.store.remove(self.test_stream,
                                self.test_data['timestamp'][350],
                                self.test_data['timestamp'][500])

        # XXX___----===XXX
        #   ^|         ^
        await self.store.remove(self.test_stream,
                                self.test_data['timestamp'][250],
                                self.test_data['timestamp'][300])

        # extract the data, should have an interval gap between 249 and 500
        # and a *single* interval boundary at 250

        extracted_data = []

        async def callback(rx_data, layout, decimated):
            extracted_data.append(rx_data)

        await self.store.extract(self.test_stream, start=None, end=None,
                                 callback=callback)
        extracted_data = np.hstack(extracted_data)

        # beginning is unchanged
        np.testing.assert_array_equal(extracted_data[:249], self.test_data[:249])
        # interval boundary marking missing data
        np.testing.assert_array_equal(extracted_data[250],
                                      pipes.interval_token(self.test_stream.layout))
        # end is unchanged (closing interval boundary ignored)
        np.testing.assert_array_equal(extracted_data[251:], self.test_data[500:])

        # two intervals of data
        intervals = await self.store.intervals(self.test_stream, start=None, end=None)
        ts = self.test_data['timestamp']
        expected = [[ts[0], ts[249] + 1],
                    [ts[500], ts[-1] + 1]]
        self.assertEqual(intervals, expected)

    async def _test_destroy(self):
        await self.store.destroy(self.test_stream)
        records = await self.store.info([self.test_stream])
        info = records[self.test_stream.id]
        self.assertEqual(info.start, None)
        self.assertEqual(info.end, None)
        self.assertEqual(info.total_time, 0)
        self.assertEqual(info.rows, 0)
        self.assertEqual(info.bytes, 0)

    async def _test_intervals(self):

        ts = self.test_data['timestamp']

        async def _insert_boundary(_conn, _ts: int):
            dt = datetime.datetime.fromtimestamp(_ts / 1e6, tz=datetime.timezone.utc)
            query = "INSERT INTO data.stream100_intervals (time) VALUES ('%s')" % dt
            await _conn.execute(query)

        conn: asyncpg.Connection = await asyncpg.connect(self.db_url)

        # XXXX|XXXXX|XXXXX
        await _insert_boundary(conn, ts[99] + 1)
        await _insert_boundary(conn, ts[199] + 1)
        expected = [[ts[0], ts[99] + 1],
                    [ts[100], ts[199] + 1],
                    [ts[200], ts[-1] + 1]]
        intervals = await self.store.intervals(self.test_stream, None, None)
        self.assertEqual(expected, intervals)

        # insert an extra boundary that does not create a new interval
        # XXXX||XXXXX|XXXXX
        await _insert_boundary(conn, ts[99] + 10)

        intervals = await self.store.intervals(self.test_stream, None, None)
        self.assertEqual(expected, intervals)

        # insert boundaries at the beginning and end of the data,
        # this should also not create any new intervals
        # |XXXX||XXXXX|XXXXX|

        await _insert_boundary(conn, ts[0] - 1)
        await _insert_boundary(conn, ts[-1] + 1)
        intervals = await self.store.intervals(self.test_stream, None, None)
        self.assertEqual(expected, intervals)

        # insert an actual boundary, this should create a new interval
        # |XXXX||XXXXX|XXX|X|
        await _insert_boundary(conn, ts[780] + 1)
        intervals = await self.store.intervals(self.test_stream, None, None)
        expected = [[ts[0], ts[99] + 1],
                    [ts[100], ts[199] + 1],
                    [ts[200], ts[780] + 1],
                    [ts[781], ts[-1] + 1]]
        self.assertEqual(expected, intervals)
        await conn.close()

    async def _test_consolidate(self):
        # intervals less than max_gap us apart are consolidated
        # data: 100 samples spaced at 1000us
        test_stream = DataStream(id=1, name="stream1", datatype=DataStream.DATATYPE.FLOAT32, keep_us=DataStream.KEEP_ALL,
                                 decimate=True, elements=[Element(name="e%d" % x) for x in range(3)])
        pipe = pipes.LocalPipe(test_stream.layout)
        nrows = 955
        orig_data = helpers.create_data(layout=test_stream.layout, length=nrows)
        chunks = [orig_data[:300], orig_data[305:400], orig_data[402:700], orig_data[800:]]
        # data: |++++++|  |+++++++++|    |++++++|    |++++|
        #               ^--5000 us    ^--2000 us   ^---0.1 sec (retained)
        chunks = [orig_data[:300], orig_data[305:400], orig_data[402:700], orig_data[800:850], orig_data[852:]]
        # data: |++++++|  |+++++++++|    |++++++|    |++++|  |++++|
        #               ^--5000 us    ^--2000 us   |        ^--- 2000 us
        #                                          `---0.1 sec (retained)
        task = await self.store.spawn_inserter(test_stream, pipe)
        for chunk in chunks:
            await pipe.write(chunk)
            await pipe.close_interval()
        await pipe.close()
        await task

        # extract data
        extracted_data = []

        rx_chunks = []

        async def callback(rx_data, layout, factor):
            #if rx_data[0] != pipes.interval_token(layout):
            if rx_data[-1][0]==0:
                rx_chunks.append(rx_data[:-1])
            else:
                rx_chunks.append(rx_data)

        await self.store.consolidate(test_stream, start=None, end=None, max_gap=6e3)
        await self.store.extract(test_stream, start=None, end=None, callback=callback)

        # should only be two intervals left (the first two are consolidated)
        np.testing.assert_array_equal(rx_chunks[0], np.hstack(chunks[:3]))
        np.testing.assert_array_equal(rx_chunks[1], np.hstack(chunks[3:]))
        self.assertEqual(len(rx_chunks), 2)

    async def _test_consolidate_with_time_bounds(self):
        # intervals less than max_gap us apart between start and end are consolidated
        # data: 100 samples spaced at 1000us
        test_stream = DataStream(id=1, name="stream1", datatype=DataStream.DATATYPE.FLOAT32, keep_us=DataStream.KEEP_ALL,
                                 decimate=True, elements=[Element(name="e%d" % x) for x in range(3)])
        pipe = pipes.LocalPipe(test_stream.layout)
        nrows = 955
        orig_data = helpers.create_data(layout=test_stream.layout, length=nrows)
        chunks = [orig_data[:300], orig_data[305:400], orig_data[402:700], orig_data[800:850],orig_data[852:]]
        # data: |++++++|  |+++++++++|    |++++++|    |++++|  |++++|
        #               ^--(retained) ^--2000 us   |        ^--- 2000 us (retained)
        #                                          `---0.1 sec (retained)
        task = await self.store.spawn_inserter(test_stream, pipe)
        for chunk in chunks:
            await pipe.write(chunk)
            await pipe.close_interval()
        await pipe.close()
        await task

        # extract data
        extracted_data = []

        rx_chunks = []

        async def callback(rx_data, layout, factor):
            if rx_data[-1][0]==0:
                rx_chunks.append(rx_data[:-1])
            else:
                rx_chunks.append(rx_data)

        await self.store.consolidate(test_stream, start=chunks[1][3]['timestamp'],
                                     end=chunks[3][3]['timestamp'], max_gap=6e3)
        await self.store.extract(test_stream, start=None, end=None, callback=callback)

        # should only be four intervals left (the inner short intervals are consolidated)
        np.testing.assert_array_equal(rx_chunks[0], chunks[0])
        np.testing.assert_array_equal(rx_chunks[1], np.hstack(chunks[1:3]))
        np.testing.assert_array_equal(rx_chunks[2], chunks[3])
        np.testing.assert_array_equal(rx_chunks[3], chunks[4])

        self.assertEqual(len(rx_chunks), 4)

    async def _test_db_info(self):
        db_info = await self.store.dbinfo()
        self.assertTrue(os.path.isdir(db_info.path))
        self.assertGreater(db_info.other, 0)
        self.assertGreater(db_info.reserved, 0)
        self.assertGreater(db_info.free, 0)
        self.assertGreater(db_info.size, 0)

    async def _test_info(self):
        # create another stream
        empty_stream = DataStream(id=103, name="empty stream", datatype=DataStream.DATATYPE.INT32,
                                  keep_us=DataStream.KEEP_ALL, decimate=True,
                                  elements=[Element(name="e%d" % x) for x in range(8)])
        stream2 = DataStream(id=104, name="stream2", datatype=DataStream.DATATYPE.INT32,
                             keep_us=DataStream.KEEP_ALL, decimate=True,
                             elements=[Element(name="e%d" % x) for x in range(8)])
        pipe = pipes.LocalPipe(stream2.layout)
        test_data = helpers.create_data(layout=stream2.layout, length=800)
        task = await self.store.spawn_inserter(stream2, pipe)
        await pipe.write(test_data)
        await pipe.close()
        await task
        records = await self.store.info([self.test_stream, stream2, empty_stream])
        # check stream1
        info = records[self.test_stream.id]
        self.assertEqual(info.start, self.test_data['timestamp'][0])
        self.assertEqual(info.end, self.test_data['timestamp'][-1])
        self.assertEqual(info.total_time, info.end - info.start)
        # rows are approximate
        self.assertLess(abs(len(self.test_data) - info.rows), len(self.test_data) * 0.1)
        self.assertGreater(info.bytes, 0)

        # check stream2
        info = records[stream2.id]
        self.assertEqual(info.start, test_data['timestamp'][0])
        self.assertEqual(info.end, test_data['timestamp'][-1])
        self.assertEqual(info.total_time, info.end - info.start)
        self.assertLess(abs(len(test_data) - info.rows), len(test_data) * 0.1)
        self.assertGreater(info.bytes, 0)

        # check the empty stream
        info = records[empty_stream.id]
        self.assertEqual(info.start, None)
        self.assertEqual(info.end, None)
        self.assertEqual(info.total_time, 0)
        self.assertEqual(info.rows, 0)
        self.assertEqual(info.bytes, 0)

    async def _test_actions_on_empty_streams(self):
        # make sure all actions can be performed on streams with no data
        empty_stream = DataStream(id=103, name="empty stream", datatype=DataStream.DATATYPE.INT32,
                                  keep_us=DataStream.KEEP_ALL, decimate=True,
                                  elements=[Element(name="e%d" % x) for x in range(8)])

        # ==== extract =====
        cb_executed = False
        cb_layout = None
        cb_factor = None

        async def callback(rx_data, layout, factor):
            nonlocal cb_executed, cb_layout, cb_factor
            cb_executed = True
            self.assertEqual(len(rx_data), 0)
            cb_layout = layout
            cb_factor = factor

        # stream with no data tables
        await self.store.extract(empty_stream, start=None, end=None, callback=callback)
        self.assertTrue(cb_executed)
        self.assertEqual(cb_layout, empty_stream.layout)
        self.assertEqual(cb_factor, 1)
        cb_executed = False
        await self.store.extract(empty_stream, start=100, end=None, callback=callback)
        self.assertTrue(cb_executed)
        self.assertEqual(cb_layout, empty_stream.layout)
        self.assertEqual(cb_factor, 1)
        cb_executed = False
        await self.store.extract(empty_stream, start=100, end=300, callback=callback)
        self.assertTrue(cb_executed)
        self.assertEqual(cb_layout, empty_stream.layout)
        self.assertEqual(cb_factor, 1)
        cb_executed = False
        await self.store.extract(empty_stream, start=None, end=100, callback=callback)
        self.assertTrue(cb_executed)
        self.assertEqual(cb_layout, empty_stream.layout)
        self.assertEqual(cb_factor, 1)
        cb_executed = False
        await self.store.extract(empty_stream, start=100, end=300, max_rows=40, callback=callback)
        self.assertTrue(cb_executed)
        self.assertEqual(cb_layout, empty_stream.layout)
        self.assertEqual(cb_factor, 1)
        cb_executed = False
        await self.store.extract(empty_stream, start=100, end=300, decimation_level=40, callback=callback)
        self.assertTrue(cb_executed)
        self.assertEqual(cb_layout, empty_stream.decimated_layout)
        self.assertEqual(cb_factor, 40)

        # non-existent decimation level
        await self.store.extract(self.test_stream, start=None, end=None, decimation_level=97,
                                 callback=callback)

        #  ==== intervals ====
        intervals = await self.store.intervals(empty_stream, None, None)
        self.assertEqual(0, len(intervals))
        intervals = await self.store.intervals(empty_stream, 100, None)
        self.assertEqual(0, len(intervals))
        intervals = await self.store.intervals(empty_stream, None, 100)
        self.assertEqual(0, len(intervals))
        intervals = await self.store.intervals(empty_stream, 100, 300)
        self.assertEqual(0, len(intervals))

        #  ==== remove ====
        await self.store.remove(empty_stream, None, None)
        await self.store.remove(empty_stream, 100, None)
        await self.store.remove(empty_stream, None, 100)
        await self.store.remove(empty_stream, 100, 300)

        # ==== destroy ====
        await self.store.destroy(empty_stream)



import sys
import tempfile
import os
import pdb

from joule.client.helpers.args import (
    module_args,
    validate_time_bounds,
    read_module_config,
    read_stream_configs)
from joule.errors import ConfigurationError
import unittest


class TestNilmdbStore(unittest.TestCase):

    def test_module_args(self):
        # if the module config is not specified, return args
        # ignore the -h help flag, the main argparse should catch it
        sys.argv = ['exec_cmd', 'arg1', 'arg2', '-h']
        arg_list = module_args()
        self.assertEqual(['arg1', 'arg2', '-h'], arg_list)
        # if the module config file is present in the args
        # parse it and add the included arguments
        with tempfile.NamedTemporaryFile() as f:
            f.write(str.encode(
                """
                [Main]
                name = test
                exec_cmd = runit.sh
                [Arguments]
                marg1 = test1
                marg2 = test2
                [Inputs]
                [Outputs]
                """))
            f.flush()
            sys.argv = ['exec_cmd', 'arg1', '--module_config', f.name]
            arg_list = module_args()
            self.assertEqual(['arg1', '--module_config', f.name,
                              '--marg1', 'test1',
                              '--marg2', 'test2'], arg_list)

    def test_validate_time_bounds(self):
        # parses time strings
        start, end = validate_time_bounds("1 hour ago", "now")
        self.assertLess(start, end)

        # handles timestamps
        start, end = validate_time_bounds("1534884541000000", None)
        self.assertIsNone(end)
        self.assertEqual(start, 1534884541000000)

        # raises error if start > end
        with self.assertRaises(ConfigurationError):
            validate_time_bounds("1 hour ago", "2 hours ago")

    def test_read_stream_configs(self):
        with tempfile.TemporaryDirectory() as dirname:
            with open(os.path.join(dirname, "stream1.conf"), 'w') as f:
                f.write(
                    """
                    [Main]
                    name = plus3
                    path = /counting
                    datatype = int32
                    keep = 5d
                    decimate = yes
                    
                    [Element1]
                    name         = Plus3
                    display_type = discrete
                    """
                )
            with open(os.path.join(dirname, "stream2.conf"), 'w') as f:
                f.write(
                    """
                    [Main]
                    name = plus5
                    path = /counting/deeper
                    datatype = int32
                    keep = all
                    decimate = yes

                    [Element1]
                    name         = Plus5
                    display_type = continuous
                    """
                )
            streams = read_stream_configs(dirname)
            self.assertEqual(streams['/counting/deeper/plus5'].elements[0].display_type, "CONTINUOUS")
            self.assertEqual(streams['/counting/plus3'].keep_us, 5 * 24 * 60 * 60 * 1e6)
            self.assertEqual(len(streams.keys()), 2)

    def test_read_module_config(self):
        # read the module config file
        with tempfile.NamedTemporaryFile() as f:
            f.write(str.encode(
                """
                [Main]
                name = test
                exec_cmd = runit.sh
                [Arguments]
               
                [Inputs]
                input1 = /test/input1
                input2 = /test/input2
                
                [Outputs]
                output1 = /test/output1
                output2 = /test/output2
                """))
            f.flush()
            config = read_module_config(f.name)
            self.assertIn('Inputs', config)
            self.assertIn('Outputs', config)
        # raises configuration error if file does not exist
        with self.assertRaises(ConfigurationError):
            read_module_config("/bad/file")
        # raises configuration error on invalid config
        with self.assertRaises(ConfigurationError):
            with tempfile.NamedTemporaryFile() as f:
                f.write(str.encode(
                    """
                    invalid config
                    """))
                f.flush()
                read_module_config(f.name)



from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
import aiohttp

import joule.controllers
from tests.controllers.helpers import MockStore
from joule import app_keys

class TestRootController(AioHTTPTestCase):

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        app[app_keys.data_store] = MockStore()
        app[app_keys.name] = "test_suite"
        return app


    async def test_index(self):
        resp: aiohttp.ClientResponse = await self.client.request("GET", "/")
        msg = await resp.text()
        self.assertEqual(resp.status, 200)
        # 'joule' is somewhere in the response string
        self.assertTrue('joule' in msg.lower())


    async def test_version(self):
        resp: aiohttp.ClientResponse = await self.client.request("GET", "/version")
        version = await resp.text()
        self.assertEqual(resp.status, 200)
        # version is some non-empty string
        self.assertTrue(len(version) > 0)
        # check JSON version
        resp: aiohttp.ClientResponse = await self.client.request("GET", "/version.json")
        data = await resp.json()
        self.assertEqual(resp.status, 200)
        # version is some non-empty string
        self.assertTrue(len(data['version']) > 0)


    async def test_dbinfo(self):
        resp: aiohttp.ClientResponse = await self.client.request("GET", "/dbinfo")
        dbinfo = await resp.json()
        self.assertEqual(resp.status, 200)
        # check one of the keys to make sure this is a dbinfo object
        self.assertTrue('path' in dbinfo)
